import numpy as np
import matplotlib.pyplot as plt
import pystan
import seaborn as sb
import pickle
import csv
from my_module import *
    
fin_nm = input('Data file name (*.csv) = ')
fout_nm = input('Output file name = ')
fout = open(fout_nm, 'w')
fout.write('Data file...{}\n'.format(fin_nm))
#
#       Read data from CSV file
#
with open(fin_nm, 'r') as fin:
    data = [v for v in csv.reader(fin)]

print('data...\n', data)

N = len(data) - 1
ID = []
X = np.empty(N, dtype = float)
Res = np.empty(N, dtype = int)
for t in range(N):
    ID.append(data[t + 1][0])
    X[t] = float(data[t + 1][1])
    Res[t] = int(data[t + 1][2])

print('Number of data = ', N)
fout.write('\nNumber of data = {}\n\n'.format(N))

for i in range(N):
    print('{0}  {1}  {2}'.format(ID[i], X[i], Res[i]))
    fout.write('{0:>5s}  {1:>10.2f}  {2:>5d}\n'.format(ID[i], X[i], Res[i]))

C1C2 = np.percentile(X, [20, 80])
print('C1C2 = ', C1C2)

def f_init():
    """
                Initial values for C, mu and sigma
    """
    return dict(sgm = C1C2[1] - C1C2[0],
                C1  = C1C2[0],
                w_C = C1C2[1] - C1C2[0])
#
#       Boundaries and initial values
#
min_sgm = 0.0                                 #   lower boundary for sigma
max_sgm = (np.amax(X) - np.amin(X)) * 2       #   upper boundary for sigma

stan_data = {'N': N, 'X':X, 'Res': Res, 'min_sgm': min_sgm, 'max_sgm': max_sgm}
#
#       Sampling by Stan
#
sm = pystan.StanModel(file = 'PM3Cat_EqC.stan')

fit = sm.sampling(data = stan_data, pars = ['C', 'C1', 'C2', 'mu', 'sgm'],
                  init = f_init,
                  n_jobs = 1)
 
print(fit)
fout.write('\nSummary...\n{}\n'.format(fit))

SmplSgm = fit['sgm']    #   Samples from the posterior distribution for sigma
SmplC = fit['C']        #   Samples from the posterior distribution for C
SmplMu = fit['mu']      #   Samples from the posterior distribution for mu
#
#       Percentile points for posterior distributions
#
SgmL025 = np.percentile(SmplSgm, 2.5)
SgmMed = np.percentile(SmplSgm, 50)
SgmU975 = np.percentile(SmplSgm, 97.5)
CL025 = np.percentile(SmplC, 2.5)
CMed = np.percentile(SmplC, 50)
CU975 = np.percentile(SmplC, 97.5)
C1Med = np.percentile(fit['C1'], 50)
C2Med = np.percentile(fit['C2'], 50)
MuL025 = np.percentile(SmplMu, 2.5)
MuMed = np.percentile(SmplMu, 50)
MuU975 = np.percentile(SmplMu, 97.5)
#
#       MAP estimates for mu, sigma and C
#
#
def CalcMAPEst(samples, a = 0.05, n_points = 10000):
    """
        Calculatte a MAP estimate from a KDE graph on [Lp, Up]
        Lp and Up are 100*a/2 and 100(1-a/2) percentile points of samples
    """
    import numpy as np
    import scipy.stats as ss
    
    Lp, Up = np.percentile(samples, [100 * a/2, 100 * (1 - a/2)])  
    coord = np.linspace(Lp, Up, n_points) 
    est_pdf = ss.gaussian_kde(samples).pdf(coord)  
    map_idx = np.argmax(est_pdf) 
    MAP_Est = coord[map_idx]       
    return MAP_Est, est_pdf[map_idx]


muMAP = CalcMAPEst(SmplMu)[0]
sgmMAP = CalcMAPEst(SmplSgm)[0]
CMAP = CalcMAPEst(SmplC)[0]
print('muMAP = {0}   sgmMAP = {1}   CMAP = {2}'.format(muMAP, sgmMAP, CMAP))
fout.write('\nMAP for mu = {0:<.2f}    95% CI for mu = [{1:<.2f}, {2:<.2f}]\n'.
           format(muMAP, MuL025, MuU975))
#
#       KDE plot for the samples from posterior distribution for mu
#
sb.kdeplot(SmplMu)
plt.plot([MuL025, MuU975], [0.0, 0.0], color = 'b', linewidth = 5, label = '95% CI')
plt.plot(muMAP, 0.0, 'go', markersize = 10, label = 'MAP')
plt.yticks([])
plt.xlabel('$\mu$', fontsize = 14)
plt.title('Posterior Distribution for $\mu$\n' +
          ('MAP = {0:<.2f}, 95% CI = [{1:<.2f}, {2:<.2f}]'.format(muMAP, MuL025, MuU975)),
          fontsize = 16)
plt.legend()
plt.tight_layout(h_pad = 5)
plt.show()
#
#       KDE plot for the samples from posterior distribution for sigma
#
fout.write('\nMap for sigma = {0:<.2f}    95% CI for sigma = [{1:<.2f}, {2:<.2f}]\n'.
           format(sgmMAP, SgmL025, SgmU975))
sb.kdeplot(SmplSgm)
plt.plot([SgmL025, SgmU975], [0.0, 0.0], color = 'b', linewidth = 5, label = '95% CI')
plt.plot(sgmMAP, 0.0, 'go', markersize = 10, label = 'MAP')
plt.yticks([])
plt.xlabel('$\sigma$', fontsize = 14)
plt.title('Posterior Distribution for $\sigma$\n' +
          ('MAP = {0:<.2f}, 95% CI = [{1:<.2f}, {2:<.2f}]'.format(sgmMAP, SgmL025, SgmU975)),
          fontsize = 16)
plt.legend()
plt.tight_layout(h_pad = 5)
plt.show()
#
#       KDE plot for the samples from posterior distribution for C
#
fout.write('\nMAP for C = {0:<.2f}      95% CI for C = [{1:<.2f}, {2:<.2f}]\n'.
           format(CMAP, CL025, CU975))
sb.kdeplot(SmplC)
plt.plot([CL025, CU975], [0.0, 0.0], color = 'b', linewidth = 5, label = '95% CI')
plt.plot(CMAP, 0.0, 'go', markersize = 10, label = 'MAP')
plt.yticks([])
plt.xlabel('C', fontsize = 14)
plt.title('Posterior Distribution for C\n' +
          ('MAP = {0:<.2f}, 95% CI = [{1:<.2f}, {2:<.2f}]'.format(CMAP, CL025, CU975)),
          fontsize = 16)
plt.legend()
plt.tight_layout(h_pad = 5)
plt.show()
#
#       Prob(z < s75) = 0.75 for the standard normal distribution
#
s75 = bisect(f_phi, 0.75, 0.5, 1.0)             #   f_phi from my_module.py
print('s75 = ', s75, '     p = ', f_phi(s75))

JND = sgmMAP * s75
PSE = muMAP
fout.write('\nMAP for PSE = {0:<.2f},    95% CI for PSE = [{1:<.2f}, {2:<.2f}]'.format(
            PSE, MuL025, MuU975))
fout.write('\nMAP for JND = {0:<.2f},    95% CI for JND= [{1:<.2f}, {2:<.2f}]'.format(
            JND, s75 * SgmL025, s75 * SgmU975))
#
#       Data and the psychometric function
#
pf = PF(muMAP, sgmMAP)          #   Class PF from my_module.py
minX = X[np.argmin(X)]
maxX = X[np.argmax(X)]
xvalues = []
yvalues = []
yvalues_L = []
yvalues_R = []
for i in range(1001):
    t = minX + i * (maxX - minX) / 1000.0
    xvalues.append(t)
    yvalues.append(pf.v(t))
    yvalues_L.append(pf.v(t + CMAP))
    yvalues_R.append(pf.v(t - CMAP))
plt.title('Psychometric Functions\n' +
          ('PSE = {0:<.2f},  JND = {1:<.2f},  C = {2:<.2f}'.format(PSE, JND, CMAP)),
          fontsize = 16)
plt.plot(xvalues, yvalues, 'k--', linewidth = 3, label = 'PF')
plt.plot(xvalues, yvalues_L, 'g-', linewidth = 3, label = 'PF(>1)')
plt.plot(xvalues, yvalues_R, 'b-', linewidth = 3, label = 'PF(>2)')
plt.yticks([0.0, 0.5, 0.75, 1.0])
#
#       Responses are scaled to values from 0 to 1, and added small fluctuations
#
Res_rn = (Res - 1) / 2 + np.random.uniform(-0.05, 0.05, len(Res))
plt.scatter(X, Res_rn, alpha = 0.5, color = 'g', label = 'Data')
plt.plot([minX, maxX], [0.0, 0.0], color = 'k')
plt.plot([minX, maxX], [1.0, 1.0], color = 'k')
plt.plot([minX, PSE, PSE], [0.5, 0.5, 0.0], color = 'k')
plt.plot([minX, PSE + JND, PSE + JND], [0.75, 0.75, 0.0], color = 'k')
plt.legend(loc = 7)
plt.xlabel('Stimulus value', fontsize = 12)
plt.ylabel('Probability', fontsize = 12)
plt.show()
#
#       Observed responses smoothed by Kernel density estimation 
#
w_1 = np.empty(len(Res))        #   Weights for Response >= 2
w_2 = np.empty(len(Res))        #   Weights for response >= 3
for t in range(len(Res)):
    w_1[t] = 0 if Res[t] < 2 else 1
    w_2[t] = 0 if Res[t] < 3 else 1

X_range = np.arange(minX, maxX, (maxX - minX) / 100)
w_k = JND / 2

while True:
    pm_1 = []           #   KDE values for Response >= 2
    pm_2 = []           #   KDE values for Response >= 3
    for v in X_range:
        pm_1.append(k_ave(v, X, w_1, w_k))
        pm_2.append(k_ave(v, X, w_2, w_k))
        
    plt.title('Observed and Estimated PFs\n' +
              ('PSE = {0:<.2f},  JND = {1:<.2f},  C = {2:<.2f}'.format(PSE, JND, CMAP)),
              fontsize = 16)
    plt.plot(xvalues, yvalues_L, 'g-', linewidth = 3, label = 'PF(>1)')
    plt.plot(xvalues, yvalues_R, 'b-', linewidth = 3, label = 'PF(>2)')
    plt.yticks([0.0, 0.5, 1.0])
    plt.plot(X_range, pm_1, 'g--', linewidth = 2, label = 'Res(>1)')
    plt.plot(X_range, pm_2, 'b--', linewidth = 2, label = 'Res(>2)')
    plt.xlabel('Stimulus value', fontsize = 12)
    plt.legend(loc = 'upper left')
    plt.show()
    print('\nThe current value of w_k = {}'.format(w_k))
    w_k = float(input('The nexr value of w_k = '))
    if w_k <= 0.0:
        break

fout.close()
print('\nOutput file {0} was saved.\n'.format(fout_nm))
